/************************************************************************\
**  ftp.c - File Transfert Protocol                                     **
**                                                                      **
**  Copyright (c) 2004 Christophe Blaess <ccb@club-internet.fr>         **
**    ---------------------------------------------------------------   **
**                                                                      **
** This program is free software; you can redistribute it and/or modify **
** it under the terms of the GNU General Public License as published by **
** the Free Software Foundation.                                        **
**                                                                      **
**  This program is distributed in the hope that it will be useful,     **
** but WITHOUT ANY WARRANTY; without even the implied warranty of       **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        **
** GNU General Public License for more details.                         **
**                                                                      **
** You should have received a copy of the GNU General Public License    **
** along with this program; if not, write to the Free Software          **
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA             **
** 02111-1307  USA                                                      **
**                                                                      **
**    ---------------------------------------------------------------   **
**                                                                      **
** Ce programme est libre, vous pouvez le redistribuer et/ou le modifier**
** selon les termes de la Licence Publique Gnrale GNU publie par la  **
** Free Software Foundation.                                            **
**                                                                      **
** Ce programme est distribu car potentiellement utile, mais SANS      **
** AUCUNE GARANTIE, ni explicite ni implicite, y compris les garanties  **
** de commercialisation ou d'adaptation dans un but spcifique.         **
** Reportez-vous  la Licence Publique Gnrale GNU pour plus de dtails**
**                                                                      **
** Vous devez avoir reu une copie de la Licence Publique Gnrale GNU  **
** en mme temps que ce programme ; si ce n'est pas le cas, crivez    **
** la Free Software Foundation, Inc, 59 Temple Place, Suite 330, Boston **
** MA 02111-1307, tats-Unis.                                           **
**                                                                      **
\************************************************************************/

	/* See : 
		+ RFC 959  - File Transfer Protocol
		= RFC 1579 - Firewall-Friendly FTP
		= RFC 1635 - How to use anonymous FTP
		= RFC 2577 - FTP security considerations
	 */

#ifndef MULTISERV_FTP_ANONYMOUS_ONLY

	#define _XOPEN_SOURCE	500
		/* needed for crypt */
	#define _BSD_SOURCE
		/* needed for strcasecmp if _XOPEN_SOURCE defined */
#endif

	#include <ctype.h>
	#include <dirent.h>
	#include <errno.h>
	#include <getopt.h>
	#include <glob.h>
	#include <signal.h>
	#include <stdarg.h>
	#include <stdio.h>
	#include <stdlib.h>
	#include <string.h>
	#include <unistd.h>
	#include <time.h>
	#include <unistd.h>
	#include <arpa/inet.h>
	#include <netinet/in.h>
	#include <sys/socket.h>
	#include <sys/types.h>
	#include <sys/stat.h>
	#include <sys/fcntl.h>

	#include "multiserv.h"
	#include "multiserv_utils.h"

	#include "ftp.h"

#if defined(MULTISERV_FTP_TCP)

	static int	ftp_dialog (int sock, struct sockaddr_in * unused_client_address);

	static int	ftp_quit  (int control_socket);
	static int	ftp_user  (int control_socket, char ** logname, const char * arg);
	static int	ftp_pass  (int control_socket, const char * logname, const char * password);
	static int 	ftp_noop  (int control_socket);
	static int	ftp_syst  (int control_socket);
	static int	ftp_pwd   (int control_socket);
	static int	ftp_cwd   (int control_socket, const char * pathname);
	static int	ftp_cdup  (int control_socket);
	static int	ftp_mkd   (int control_socket, const char * directory);
	static int	ftp_rmd   (int control_socket, const char * directory);
	static int	ftp_dele  (int control_socket, const char * file);
	static int	ftp_rnfr  (int control_socket, const char * file, char ** rename);
	static int	ftp_rnto  (int control_socket, const char * file, char ** rename);
	static int	ftp_pasv  (int control_socket, int * data_socket);
	static int	ftp_port  (int control_socket, int * data_socket, const char * string);
	static int	ftp_help  (int control_socket);
	static int	ftp_list  (int control_socket, int * data_socket, const char * arg, int long_format);
	static int	ftp_list_dir_entry (const char * dir, const char * name, int data_socket, int long_format);
	static int	ftp_retr  (int control_socket, int * data_socket, const char * arg, char type);
	static int	ftp_stor  (int control_socket, int * data_socket, const char * arg, char type, char * openmode);
	static int	ftp_file_attribut (int control_socket, const char * filename, const char * attrib);

	static int	ftp_mode  (int control_socket, const char * arg, char * mode);
	static int	ftp_type  (int control_socket, const char * arg, char * type);
	static int	ftp_stru  (int control_socket, const char * arg, char * stru);
	
	static char *	ftp_getcwd (void);
	static int	ftp_check_data_socket (int control_socket, int * data_socket, int passive);

	static int	ftp_send_string (int sock, const char * string, char type);
	static int	ftp_send_block  (int sock, const char * string, size_t size, char type);
	static int	ftp_wait_string (int sock);
	static int	ftp_get_block   (int sock, char * block, size_t size, char type);

#ifndef MULTISERV_FTP_ANONYMOUS_ONLY
	static int	ftp_check_access(const char * logname, const char * password);
#endif

/***********************************************************************\
			 Public functions
\***********************************************************************/
	
	int
ftp_description (FILE * stream)
{
	return (fprintf(stream, "File Transfert Protocol (FTP) rfc.959 (TCP/IP)"));
}


	int
ftp_main (int argc, char * argv[])
{
	multiserv_log(MULTISERV_LOG_MESSAGE, "ftp", "TCP server start");
	return (tcp_server(ftp_dialog));
}


/***********************************************************************\
			 Private functions
\***********************************************************************/

	static int
ftp_dialog (int control_socket, struct sockaddr_in * client_address)
{
	/* Checked: 1 */

	char *	arg = NULL;
	char *	logname = NULL;
	char *	rename_file = NULL;
	int	access_granted = 0;
	struct	sockaddr_in destination;
	int	data_socket = -1;
	int	passive = 0;

	char	type = 'a';
	char	stru = 'f';
	char	mode = 's';
	
	multiserv_debug("ftp", "Got a connection, sending banner");
	#ifdef MULTISERV_FTP_VERBOSE
		ftp_send_string(control_socket, "220 Multiserv FTP server ready\n", 'a');
	#else
		ftp_send_string(control_socket, "220\n", 'a');
	#endif	
	memset(&destination, 0, sizeof(struct sockaddr_in));
	
	while(1) {
	
		#ifdef MULTISERV_FTP_TIMEOUT
			if (MULTISERV_FTP_TIMEOUT > 0)
				alarm(MULTISERV_FTP_TIMEOUT);
		#endif
		if (ftp_wait_string(control_socket) != 0) {
			if (errno == EINTR) {
			#ifdef MULTISERV_FTP_VERBOSE
				ftp_send_string(control_socket, "321 Timeout!\n", 'a');
			#else
				ftp_send_string(control_socket, "221\n", 'a');
			#endif
			}
			break;
		}
		#ifdef MULTISERV_FTP_TIMEOUT
			alarm(0);
		#endif

		if ((rename_file != NULL)
		 && (strncasecmp(multiserv_buffer, "RNTO ", 5) != 0)) {
			free(rename_file);
			rename_file = NULL;
		}

		arg = multiserv_buffer;
		while ((arg[0] != '\0') && (arg[0] != ' '))
			arg ++;
		while(arg[0] == ' ')
			arg ++;

		multiserv_debug("ftp",multiserv_buffer);

		/* QUIT is always available */
		if (strncasecmp (multiserv_buffer, "QUIT", 4) == 0) {
			ftp_quit(control_socket);
			break;
		}

		/* USER / PASS is only allowed when access is NOT already granted,
		   because we set UID/GID and even chroot during anonymous connections */
		if ((strncasecmp (multiserv_buffer, "USER ", 5) == 0)
		 && (! access_granted)) {
			if (ftp_user(control_socket, & logname, arg) != 0)
				break;
			continue;
		}

		if ((strncasecmp(multiserv_buffer, "PASS ", 5) == 0)
		 && (! access_granted)) {
			access_granted = (ftp_pass(control_socket, logname, arg) == 0);
			memset(multiserv_buffer, 0, MULTISERV_BUFFER_LEN); /* paranoia */
			continue;
		}

		/* All other commands need preliminary login */
		if (! access_granted) {
			#ifdef MULTISERV_FTP_VERBOSE
				ftp_send_string(control_socket, "530 You need to log in first!\n", 'a');
			#else
				ftp_send_string(control_socket, "530\n", 'a');
			#endif
			continue;
		}

		if (strcasecmp(multiserv_buffer, "SYST") == 0) {
			if (ftp_syst(control_socket) != 0)
				break;
			continue;
		}
				
		if (strcasecmp(multiserv_buffer, "HELP") == 0) {
			if (ftp_help(control_socket) != 0)
				break;
			continue;
		}
		
		if ((strcasecmp(multiserv_buffer, "NOOP") == 0) 
		||  (strncasecmp(multiserv_buffer, "ALLO", 4) == 0)) {
			if (ftp_noop(control_socket) != 0)
				break;
			continue;
		}
				
		if ((strcasecmp(multiserv_buffer, "PWD") == 0)
		 || (strcasecmp(multiserv_buffer, "XPWD") == 0)) {
			if (ftp_pwd(control_socket) != 0)
				break;
			continue;
		}

		if ((strncasecmp(multiserv_buffer, "CWD ", 4) == 0)
		 || (strncasecmp(multiserv_buffer, "XCWD ", 5) == 0)) {
		 	if (ftp_cwd(control_socket, arg) != 0)
				break;
			continue;
		}

		if ((strcasecmp(multiserv_buffer, "CDUP") == 0)
		|| (strcasecmp(multiserv_buffer, "XCUP") == 0)) {
			if (ftp_cdup(control_socket) != 0)
				break;
			continue;
		}

		if ((strncasecmp(multiserv_buffer, "MKD ", 4) == 0)
		 || (strncasecmp(multiserv_buffer, "XMKD ", 5) == 0)) {
		 	if (ftp_mkd(control_socket, arg) != 0)
				break;
			continue;
		}


		if ((strncasecmp(multiserv_buffer, "RMD ", 4) == 0)
		 || (strncasecmp(multiserv_buffer, "XRMD ", 5) == 0)) {
		 	if (ftp_rmd(control_socket, arg) != 0)
				break;
			continue;
		}
		
		if (strncasecmp(multiserv_buffer, "DELE ", 5) == 0) {
			if (ftp_dele(control_socket, arg) != 0)
				break;
			continue;
		}
		
		if (strncasecmp(multiserv_buffer, "RNFR ", 5) == 0) {
			if (ftp_rnfr(control_socket, arg, & rename_file) != 0)
				break;
			continue;
		}
		
		if (strncasecmp(multiserv_buffer, "RNTO ", 5) == 0) {
			if (ftp_rnto(control_socket, arg, & rename_file) != 0)
				break;
			continue;
		}
		
		if (strncasecmp(multiserv_buffer, "SIZE ", 5) == 0) {
			if (ftp_file_attribut(control_socket, arg, "size") != 0)
				break;
			continue;
		}

		if (strncasecmp(multiserv_buffer, "MDTM ", 5) == 0) {
			if (ftp_file_attribut(control_socket, arg, "mtime") != 0)
				break;
			continue;
		}

		if (strncasecmp(multiserv_buffer, "PORT ", 5) == 0) {
			passive = 0;
			if (ftp_port(control_socket, & data_socket, arg) != 0)
				break;
			continue;
		}

		if (strcasecmp(multiserv_buffer, "PASV") == 0) {
			passive = 1;
			if (ftp_pasv(control_socket, & data_socket) != 0)
				break;
			continue;
		}
		
		if (strncasecmp(multiserv_buffer,"LIST", 4) == 0) {
			if (ftp_check_data_socket(control_socket, & data_socket, passive) != 0)
				continue;
			if (ftp_list(control_socket, & data_socket, arg, 1) != 0)
				break;
			continue;
		}

		if (strncasecmp(multiserv_buffer,"NLST", 4) == 0) {
			if (ftp_check_data_socket(control_socket, & data_socket, passive) != 0)
				continue;
			if (ftp_list(control_socket, & data_socket, arg, 0) != 0)
				break;
			continue;
		}
		
		if (strncasecmp(multiserv_buffer, "TYPE ", 5) == 0) {
			if (ftp_type(control_socket, arg, & type) != 0)
				break;
			continue;
		}

		if (strncasecmp(multiserv_buffer, "MODE ", 5) == 0) {
			if (ftp_mode(control_socket, arg, & mode) != 0)
				break;
			continue;
		}

		if (strncasecmp(multiserv_buffer, "STRU ", 5) == 0) {
			if (ftp_stru(control_socket, arg, & stru) != 0)
				break;
			continue;
		}

		if (strncasecmp(multiserv_buffer, "RETR ", 5) == 0) {
			if (ftp_check_data_socket(control_socket, & data_socket, passive) != 0)
				continue;
			if (ftp_retr(control_socket, & data_socket, arg, type) != 0)
				break;
			continue;
		}

		if (strncasecmp(multiserv_buffer, "STOR ", 5) == 0) {
			if (ftp_check_data_socket(control_socket, & data_socket, passive) != 0)
				continue;
			if (ftp_stor(control_socket, & data_socket, arg, type, "w") != 0)
				break;
			continue;
		}

		if (strcasecmp(multiserv_buffer, "STOU") == 0) {
			if (ftp_check_data_socket(control_socket, & data_socket, passive) != 0)
				continue;
			if (ftp_stor(control_socket, & data_socket, arg, type, "x") != 0)
				break;
			continue;
		}

		if (strncasecmp(multiserv_buffer, "APPE ", 5) == 0) {
			if (ftp_check_data_socket(control_socket, & data_socket, passive) != 0)
				continue;
			if (ftp_stor(control_socket, & data_socket, arg, type, "a") != 0)
				break;
			continue;
		}

		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "500 Command unknown\n", 'a');
		#else
			ftp_send_string(control_socket, "500\n", 'a');
		#endif
		multiserv_debug("ftp", "Unknown command");
	}
	close(control_socket);
	if (logname != NULL)
		free(logname);
	logname = NULL;
	if (rename_file != NULL)
		free(rename_file);
	rename_file = NULL;
	return (0);
}


/* ------------------- FTP Commands implementation ------------------ */

	static int
ftp_quit(int control_socket)
{
	/* Checked: 1 */
	#ifdef MULTISERV_FTP_VERBOSE
		return ftp_send_string(control_socket, "221 Bye.\n", 'a');
	#else
		return ftp_send_string(control_socket, "221\n", 'a');
	#endif
}

		
	static int
ftp_user(int control_socket, char ** logname, const char * arg)
{
	/* Checked : 1 */
	if (logname == NULL) {
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 Local error (USER1)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		return (-1);
	}
	/* While access is still not granted, multiple successive
	   USER commands are allowed */
	if (* logname != NULL)
		free (* logname);
	* logname = malloc (strlen(arg) + 1);
	if (* logname == NULL) {
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 Local error (USER2)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		return (-1);
	}
	strcpy(* logname, arg);
	#ifdef MULTISERV_FTP_ANONYMOUS
		if (strcmp(* logname, "anonymous") == 0)
			#ifdef MULTISERV_FTP_VERBOSE
				return ftp_send_string(control_socket, "331 Please send your e-mail address as password.\n", 'a');
			#else
				return ftp_send_string(control_socket, "331\n", 'a');
			#endif
	#endif
	#ifdef MULTISERV_FTP_VERBOSE
		return ftp_send_string(control_socket, "331 User name ok, need password.\n", 'a');
	#else
		return ftp_send_string(control_socket, "331\n", 'a');
	#endif
}


	static int
ftp_pass(int control_socket, const char * logname, const char * password)
{
	/* Checked: 1 */
	if (logname == NULL) {
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "503 Please log in with USER first\n", 'a');
		#else
			ftp_send_string(control_socket, "503\n", 'a');
		#endif
		return (-1);
	}
	if (strcmp(logname, "anonymous") == 0) {
		#ifdef MULTISERV_FTP_ANONYMOUS
			/* Note that anonymous ftp is only allowed for server
			  ran by root (chroot) */
			if ((chdir(MULTISERV_FTP_ANONYMOUS_DIR) != 0)
			 || (chroot(MULTISERV_FTP_ANONYMOUS_DIR) != 0)
			 || (setgid(MULTISERV_FTP_ANONYMOUS_GID) != 0)
			 || (setuid(MULTISERV_FTP_ANONYMOUS_UID) != 0)) {
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send_string(control_socket, "421 Service not available\n", 'a');
				#else
					ftp_send_string(control_socket, "421\n", 'a');
				#endif
				return (-1);
			}
			#ifdef MULTISERV_FTP_VERBOSE
				ftp_send_string(control_socket, "230 Welcome A.N.Onymous.\n", 'a');
			#else
				ftp_send_string(control_socket, "230\n", 'a');
			#endif
			/* ## TODO: log connection */
			return (0);
		#endif
	}
	#ifdef MULTISERV_FTP_ANONYMOUS_ONLY
		ftp_send_string(control_socket, "530 Login incorrect\n", 'a');
		return (-1);
	#else
	
		if (ftp_check_access(logname, password) == 0) {
			#ifdef MULTISERV_FTP_VERBOSE
				return ftp_send_string(control_socket, "230 User logged in. Welcome.\n", 'a');
			#else
				return ftp_send_string(control_socket, "230\n", 'a');
			#endif
		}
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "530 Login incorrect\n", 'a');
		#else
			ftp_send_string(control_socket, "530\n", 'a');
		#endif
		return (-1);
	#endif
}

			
	static int
ftp_syst(int control_socket)
{
	/* Checked: 1 */
	return ftp_send_string(control_socket, "215 LINUX Type: L8\n", 'a');
}


	static int
ftp_help(int control_socket)
{
	/* Checked: 1 */
	ftp_send_string(control_socket, "214- IP-multiserv FTP server for small systems\n", 'a');
	ftp_send_string(control_socket, "214- (c) 2004 - Christophe Blaess.\n", 'a');
	ftp_send_string(control_socket, "214- \n", 'a');
	ftp_send_string(control_socket, "214- The following commands are implemented.\n", 'a');
	ftp_send_string(control_socket, "214-  QUIT USER PASS SYST HELP NOOP\n", 'a');
	ftp_send_string(control_socket, "214-  ALLO  PWD XPWD  CWD XCWD CDUP\n", 'a');
	ftp_send_string(control_socket, "214-  XCUP  MKD XMKD  RMD XRMD DELE\n", 'a');
	ftp_send_string(control_socket, "214-  RNFR RNTO SIZE MDTM PORT PASV\n", 'a');
	ftp_send_string(control_socket, "214-  LIST NLST TYPE MODE STRU RETR\n", 'a');
	ftp_send_string(control_socket, "214-  STOR STOU APPE\n", 'a');
	ftp_send_string(control_socket, "214 \n", 'a');
	return(0);
}


	static int 
ftp_noop(int control_socket)
{
	/* Checked: 1 */
	/* 
	 * This routine implements NOOP but also ALLO commands
	 */
	#ifdef MULTISERV_FTP_VERBOSE
		return ftp_send_string(control_socket, "200 Ok\n", 'a');
	#else
		return ftp_send_string(control_socket, "200\n", 'a');
	#endif
}

	
	static int
ftp_pwd(int control_socket)
{
	/* Checked: 1 */
	char * pathname = ftp_getcwd();
	if (pathname == NULL) {
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 Local error (#PWD1)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		return (-1);
	}
	if (snprintf(multiserv_buffer, MULTISERV_BUFFER_LEN, "257 \"%s\"\n", pathname) >= MULTISERV_BUFFER_LEN) {
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 Local error (#PWD2)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		free(pathname);
		pathname = NULL;
		return (-1);
	}
	free(pathname);
	pathname = NULL;
	return ftp_send_string(control_socket, multiserv_buffer, 'a');
}


	static int
ftp_cwd(int control_socket, const char * pathname)
{
	/* Checked: 1 */
	if (chdir(pathname) != 0)
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "550 Unable to change directory\n", 'a');
		#else
			return ftp_send_string(control_socket, "550\n", 'a');
		#endif
	#ifdef MULTISERV_FTP_VERBOSE
		return ftp_send_string(control_socket, "250 Directory Ok\n", 'a');
	#else
		return ftp_send_string(control_socket, "250\n", 'a');
	#endif
}


	static int
ftp_cdup(int control_socket)
{
	/* Checked: 1 */
	if (chdir("..") != 0)
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "550 Unable to change directory\n", 'a');
		#else
			return ftp_send_string(control_socket, "550\n", 'a');
		#endif
	/* There's an incoherency in RFC 959: on page 50 the positive reply code for CDUP is 200,
	  but it is also said (page 62) that CDUP does have the same reply codes than CWD,
	  which uses 250. I've checked some public FTP servers and, they seems to follow the
	  second case (250).
	 */
	#ifdef MULTISERV_FTP_VERBOSE
		return ftp_send_string(control_socket, "250 Directory Ok\n", 'a');
	#else
		return ftp_send_string(control_socket, "250\n", 'a');
	#endif
}


	static int
ftp_mkd(int control_socket, const char * directory)
{
	/* Checked: 1 */
	int fd;
	char * pathname = NULL;
	if (mkdir(directory, 0755) != 0) {
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "550 Creation failed\n", 'a');
		#else
			return ftp_send_string(control_socket, "550\n", 'a');
		#endif
	}
	/* save current directory in a file descriptor */
	if ((fd = open (".", O_RDONLY)) < 0) {
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 Local error (#MKD1)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		return (-1);
	}
	if ((chdir(directory) != 0)
	 || ((pathname = ftp_getcwd()) == NULL)) {
	 	fchdir(fd);
	 	close(fd);
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 Local error (#MKD2)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		return (-1);
	}
	if (snprintf(multiserv_buffer, MULTISERV_BUFFER_LEN, "257 \"%s\"\n", pathname) >= MULTISERV_BUFFER_LEN) { 
		fchdir(fd);
		close(fd);
		free(pathname);
		pathname = NULL;
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 Local error (#MKD3)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		return (-1);
	}
	fchdir(fd);
	close(fd);
	free(pathname);
	pathname = NULL;
	return ftp_send_string(control_socket, multiserv_buffer, 'a');
}


	static int
ftp_rmd(int control_socket, const char * directory)
{
	/* Checked: 1 */
	if (rmdir(directory) != 0)
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "550 Delete failed\n", 'a');
		#else
			return ftp_send_string(control_socket, "550\n", 'a');
		#endif
	#ifdef MULTISERV_FTP_VERBOSE
		return ftp_send_string(control_socket, "250 Delete ok\n", 'a');
	#else
		return ftp_send_string(control_socket, "250\n", 'a');
	#endif
}


	static int
ftp_dele(int control_socket, const char * file)
{
	/* Checked: 1 */
	if (unlink(file) != 0)
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "550 Delete failed\n", 'a');
		#else
			return ftp_send_string(control_socket, "550\n", 'a');
		#endif
	#ifdef MULTISERV_FTP_VERBOSE
		return ftp_send_string(control_socket, "250 Delete ok\n", 'a');
	#else
		return ftp_send_string(control_socket, "250\n", 'a');
	#endif
}


	static int
ftp_rnfr(int control_socket, const char * file, char ** rename_file)
{
	/* Checked: 0 */
	struct stat filestat;
	if (lstat(file, & filestat) < 0) {
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "550 No such file or directory\n", 'a');
		#else
			return ftp_send_string(control_socket, "550\n", 'a');
		#endif
	}
	if (* rename_file != NULL)
		free(* rename_file);
	* rename_file = malloc(strlen(file) + 1);
	if (* rename_file == NULL) {
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 Local error (#RNFR1)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		return (-1);
	}
	strcpy(* rename_file, file);
	#ifdef MULTISERV_FTP_VERBOSE
		ftp_send_string(control_socket, "350 Ready to rename file\n", 'a');
	#else
		ftp_send_string(control_socket, "350\n", 'a');
	#endif
	return (0);
}


	static int
ftp_rnto(int control_socket, const char * file, char **rename_file)
{
	if (* rename_file == NULL) {
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "503 Bad sequence of commands\n", 'a');
		#else
			return ftp_send_string(control_socket, "503\n", 'a');
		#endif
	}
	if (rename(* rename_file, file) != 0) {
		free(* rename_file);
		* rename_file = NULL;
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "553 Renaming failed\n", 'a');
		#else
			return ftp_send_string(control_socket, "553\n", 'a');
		#endif
	}
	free(* rename_file);
	* rename_file = NULL;
	#ifdef MULTISERV_FTP_VERBOSE
		return ftp_send_string(control_socket, "250 Renaming ok\n", 'a');
	#else
		return ftp_send_string(control_socket, "250\n", 'a');
	#endif
}


	static int
ftp_file_attribut(int control_socket, const char * filename, const char * attrib)
{
	/* Checked: 1 */
	/* This routine implements SIZE and MDTM commands */
	struct stat filestat;
	struct tm * tm;

	if (lstat(filename, & filestat) < 0) {
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "550 No such file\n", 'a');
		#else
			return ftp_send_string(control_socket, "550\n", 'a');
		#endif
	}
	if (! S_ISREG(filestat.st_mode)) {
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "550 Not a regular file\n", 'a');
		#else
			return ftp_send_string(control_socket, "550\n", 'a');
		#endif
	}
	
	if (strcmp(attrib,"mtime") == 0) {
		tm = localtime(& filestat.st_mtime);
		snprintf(multiserv_buffer, MULTISERV_BUFFER_LEN,
			"213 %04d%02d%02d%02d%02d%02d\n",
			tm->tm_year+1900,
			tm->tm_mon + 1,
			tm->tm_mday,
			tm->tm_hour,
			tm->tm_min,
			tm->tm_sec);
		return ftp_send_string(control_socket, multiserv_buffer, 'a');
	}
	
	if (strcmp(attrib,"size") == 0) {
		snprintf(multiserv_buffer, MULTISERV_BUFFER_LEN,
			"213 %ld\n", filestat.st_size);
		return ftp_send_string(control_socket, multiserv_buffer, 'a');
	}
	#ifdef MULTISERV_FTP_VERBOSE
		ftp_send_string(control_socket, "451 Local error (#ATTRIB1)\n", 'a');
	#else
		ftp_send_string(control_socket, "451\n", 'a');
	#endif
	return (-1);
}


	static int
ftp_port(int control_socket, int * data_socket, const char * string)
{
	/* Checked: 1 */
	struct sockaddr_in address;
	int a1, a2, a3, a4;
	int p1, p2;

	/* Close previously open data socket if any */
	if (* data_socket != -1)
		close (* data_socket);
	* data_socket = -1;
	
	if ((sscanf(string, "%d,%d,%d,%d,%d,%d", & a1, & a2, & a3, & a4, & p1, & p2) != 6)
	 || (a1<0)   || (a2<0)   || (a3<0)   || (a4<0)
	 || (a1>255) || (a2>255) || (a3>255) || (a4>255)) {
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "500 Illegal PORT command\n", 'a');
		#else
			ftp_send_string(control_socket, "500\n", 'a');
		#endif
		return (0);
	}

	* data_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (* data_socket < 0) {
		* data_socket = -1;
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 local error (#PORT1)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		return (-1);
	}	
	memset(& address, 0, sizeof(struct sockaddr_in));
	address.sin_family = AF_INET;
	address.sin_port = htons(p1*256 + p2);
	address.sin_addr.s_addr = htonl(a1*16777216 + a2*65536 + a3*256 + a4);

	if (connect(* data_socket, (struct sockaddr *) & address, sizeof(struct sockaddr_in)) < 0) {	
		close (* data_socket);
		* data_socket = -1;
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "425 Can't open data connection\n", 'a');
		#else
			return ftp_send_string(control_socket, "425\n", 'a');
		#endif
	}
	#ifdef MULTISERV_FTP_VERBOSE
		return ftp_send_string(control_socket, "200 PORT command Ok\n", 'a');
	#else
		return ftp_send_string(control_socket, "200\n", 'a');
	#endif
}


	static int
ftp_pasv(int control_socket, int * data_socket)
{
	/* Checked: 1 */
	/* 
	 * This routine initialize a passive data socket, using the same IP
	 * address as the control socket and a random port (provided by the
	 * kernel).
	 * The data_socket argument is filled in return with the new socket
	 * that has to be accept()ed before real data transfert.
	 */
	struct sockaddr_in address;
	socklen_t namelen;
	
	if (* data_socket != -1)
		close(* data_socket);
	* data_socket = -1;
	
	* data_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (* data_socket < 0) {
		* data_socket = -1;
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 Local error (#PASV1)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		return (-1);
	}
	namelen = sizeof(struct sockaddr_in);
	if (getsockname(control_socket, (struct sockaddr *) & address, & namelen) < 0) {
		close (* data_socket);
		* data_socket = -1;
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 local error (#PASV2)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		return (-1);
	}
	/* The kernel will provide a free port */
	address.sin_port = 0;
	if (bind(* data_socket, (struct sockaddr *) & address, sizeof(struct sockaddr_in)) < 0) {
		close (* data_socket);
		* data_socket = -1;
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 local error (#PASV3)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		return (-1);
	}
	if (listen(* data_socket, 5) != 0) {
		close (* data_socket);
		* data_socket = -1;
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 Local error (#PASV4)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		return (-1);
	}
	/* now we need to know the port of the data socket */
	namelen = sizeof(struct sockaddr_in);
	if (getsockname(* data_socket, (struct sockaddr *) & address, & namelen) < 0) {
		close (* data_socket);
		* data_socket = -1;
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 local error (#PASV5)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		return (-1);
	}
	if (snprintf(multiserv_buffer, MULTISERV_BUFFER_LEN, "227 Entering Passive Mode (%d,%d,%d,%d,%d,%d)\n",
		(ntohl(address.sin_addr.s_addr) >> 24) & 0xFF,
		(ntohl(address.sin_addr.s_addr) >> 16) & 0xFF,
		(ntohl(address.sin_addr.s_addr) >> 8) & 0xFF,
		(ntohl(address.sin_addr.s_addr)) & 0xFF,
		(ntohs(address.sin_port) >> 8) & 0xFF,
		(ntohs(address.sin_port)) & 0xFF) >= MULTISERV_BUFFER_LEN) {
		close (* data_socket);
		* data_socket = -1;
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 local error (#PASV6)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		return (-1);
	}
	if (ftp_send_string(control_socket, multiserv_buffer, 'a')  != 0) {
		close (* data_socket);
		* data_socket = -1;
		return (-1);
	}
	return (0);
}


	static int
ftp_list(int control_socket, int * data_socket, const char * arg, int long_format)
{
	/* Checked: 1 */
	char * directory = NULL;
	DIR * dp;
	struct dirent * entry;
	int option_a = 0;
	glob_t gt;
	int i;

	if (arg[0] == '-') {
		if (strchr(arg, 'l') != NULL)
			long_format = 1;
		if (strchr(arg, 'a') != NULL)
			option_a = 1;
	}
	
	if ((arg[0] == '-') || (arg[0] == '\0')) {
		directory = malloc(2);
		if (directory == NULL) {
			#ifdef MULTISERV_FTP_VERBOSE
				ftp_send_string(control_socket, "451 Local error (#LIST1)\n", 'a');
			#else
				ftp_send_string(control_socket, "451\n", 'a');
			#endif
			return (-1);
		}
		strcpy(directory, ".");
	} else {		
		directory = malloc(strlen(arg)+1);
		if (directory == NULL) {
			#ifdef MULTISERV_FTP_VERBOSE
				ftp_send_string(control_socket, "451 Local error (#LIST2)\n", 'a');
			#else
				ftp_send_string(control_socket, "451\n", 'a');
			#endif
			return (-1);
		}
		strcpy(directory, arg);
	}
	
	
	/*
	 * The entries are not sorted. I thought using scandir() and alphasort()
	 * but this functions are not in SUSv3. Do we really need to sort the
	 * entries ? RFC 959 does not say so.
	 */
	dp = opendir(directory);
	if (dp != NULL) {
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "150 Here comes the directory listing\n", 'a');
		#else
			ftp_send_string(control_socket, "150\n", 'a');
		#endif
		while ((entry = readdir(dp)) != NULL) {
			if ((entry->d_name[0] == '.') && (! option_a ))
			 	continue;
			if (ftp_list_dir_entry(directory, entry->d_name, * data_socket, long_format) != 0) {
				closedir(dp);
				free(directory);
				close(* data_socket);
				* data_socket = -1;
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send_string(control_socket, "451 Local error (#LIST3)\n", 'a');
				#else
					ftp_send_string(control_socket, "451\n", 'a');
				#endif
				return (-1);
			}
		}
		closedir(dp);
		free(directory);
		close(* data_socket);
		* data_socket = -1;
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "226 Data transfert complete\n", 'a');
		#else
			return ftp_send_string(control_socket, "226\n", 'a');
		#endif
	}
	#ifdef MULTISERV_FTP_VERBOSE
		ftp_send_string(control_socket, "150 Here comes the list\n", 'a');
	#else
		ftp_send_string(control_socket, "150\n", 'a');
	#endif
	/*
	 * If the argument is not a single directory, we assume it is a shell
	 * pattern and we perform glob()ing on it
	 */
	if (glob(directory, 0, NULL, & gt) == 0) {
		for (i = 0; i < gt.gl_pathc; i ++) {
			if (ftp_list_dir_entry(NULL, gt.gl_pathv[i], * data_socket, long_format) != 0) {
				globfree(& gt);
				free(directory);
				close(* data_socket);
				* data_socket = -1;
				#ifdef MULTISERV_FTP_VERBOSE
					ftp_send_string(control_socket, "451 Local error (#LIST4)\n", 'a');
				#else
					ftp_send_string(control_socket, "451\n", 'a');
				#endif
				return (-1);
			}
		}
	}

	globfree(& gt);
	free(directory);
	close(* data_socket);
	* data_socket = -1;
	#ifdef MULTISERV_FTP_VERBOSE
		return ftp_send_string(control_socket, "226 Data transfert complete\n", 'a');
	#else
		return ftp_send_string(control_socket, "226\n", 'a');
	#endif
}


	static int
ftp_list_dir_entry(const char * dir, const char * name, int data_socket, int long_format)
{
	/* Checked: 1 */
	char file_mode[11];
	char file_time[13];
	struct tm * tm;
	struct stat filestat;

	time_t	now = time(NULL);

	int year = localtime(& now)->tm_year;

	if (! long_format) {
		if (snprintf(multiserv_buffer, MULTISERV_BUFFER_LEN, "%s\n",name) >= MULTISERV_BUFFER_LEN)
			return (-1);
		return ftp_send_string(data_socket, multiserv_buffer, 'a');
	}
	
	if (dir != NULL) {
		if (snprintf(multiserv_buffer, MULTISERV_BUFFER_LEN, "%s/%s", dir, name) >= MULTISERV_BUFFER_LEN)
			return (-1);
	} else {
		if (snprintf(multiserv_buffer, MULTISERV_BUFFER_LEN, "%s", name) >= MULTISERV_BUFFER_LEN)
			return (-1);
	}
	if (lstat(multiserv_buffer, & filestat) < 0)
		/* Do we continue with other files (0) or halt
		   immediatly (-1) ? */
		return (-1);

	tm = localtime(& filestat.st_mtime);

	file_mode[0] = '?';
	if (S_ISBLK(filestat.st_mode))  file_mode[0]='b';
	if (S_ISCHR(filestat.st_mode))  file_mode[0]='c';
	if (S_ISDIR(filestat.st_mode))  file_mode[0]='d';
	if (S_ISFIFO(filestat.st_mode)) file_mode[0]='p';
	if (S_ISREG(filestat.st_mode))  file_mode[0]='-';
	if (S_ISLNK(filestat.st_mode))  file_mode[0]='l';
	if (S_ISSOCK(filestat.st_mode)) file_mode[0]='s';

	if (filestat.st_mode & S_IRUSR) file_mode[1]='r'; else file_mode[1] = '-';
	if (filestat.st_mode & S_IWUSR) file_mode[2]='w'; else file_mode[2] = '-';
	if (filestat.st_mode & S_IXUSR) file_mode[3]='x'; else file_mode[3] = '-';
	if (filestat.st_mode & S_IRGRP) file_mode[4]='r'; else file_mode[4] = '-';
	if (filestat.st_mode & S_IWGRP) file_mode[5]='w'; else file_mode[5] = '-';
	if (filestat.st_mode & S_IXGRP) file_mode[6]='x'; else file_mode[6] = '-';
	if (filestat.st_mode & S_IROTH) file_mode[7]='r'; else file_mode[7] = '-';
	if (filestat.st_mode & S_IWOTH) file_mode[8]='w'; else file_mode[8] = '-';
	if (filestat.st_mode & S_IXOTH) file_mode[9]='x'; else file_mode[9] = '-';
	file_mode[10] = '\0';

	if (tm->tm_year == year) {
		if (strftime(file_time, 13, "%b %2d  %Y", tm) == 0)
			return(-1);
	} else {
		if (strftime(file_time, 13, "%b %2d %02H:%02M", tm) == 0)
			return (-1);
	}
	if (snprintf(multiserv_buffer,
			MULTISERV_BUFFER_LEN,
			"%s %3d %4d %4d %9ld %s %s\n",
			file_mode,
			filestat.st_nlink,
			filestat.st_uid,
			filestat.st_gid,
			filestat.st_size,
			file_time,
			name) >= MULTISERV_BUFFER_LEN)
		return (-1);

	return ftp_send_string(data_socket, multiserv_buffer, 'a');
}


	static int
ftp_mode(int control_socket, const char * arg, char * mode)
{
	/* Checked: 1 */
	/* ## TODO: implement other modes */
	if (strcasecmp(arg, "S") == 0)
		* mode = 's';
	else if ((strcasecmp(arg, "B") == 0) || (strcasecmp(arg, "C") == 0)) {
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "504 Mode not implemented\n", 'a');
		#else
			return ftp_send_string(control_socket, "504\n", 'a');
		#endif
	} else {
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "500 Illegal MODE command\n", 'a');
		#else
			return ftp_send_string(control_socket, "500\n", 'a');
		#endif
	}
	#ifdef MULTISERV_FTP_VERBOSE
		return ftp_send_string(control_socket, "200 Mode set ok\n", 'a');
	#else
		return ftp_send_string(control_socket, "200\n", 'a');
	#endif
}


	static int
ftp_type(int control_socket, const char * arg, char * type)
{
	/* Checked: 1 */
	/* ## TODO: implement other types */
	if (strcasecmp(arg, "A") == 0)
		* type = 'a';
	else if (strcasecmp(arg,"I") == 0)
		* type = 'i';
	else if ((strcasecmp(arg, "L") == 0) || (strcasecmp(arg, "E") == 0)) {
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "504 Type not implemented\n", 'a');
		#else
			return ftp_send_string(control_socket, "504\n", 'a');
		#endif
	} else {
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "500 Illegal TYPE command\n", 'a');
		#else
			return ftp_send_string(control_socket, "500\n", 'a');
		#endif
	}
	#ifdef MULTISERV_FTP_VERBOSE
		return ftp_send_string(control_socket, "200 Type set ok\n", 'a');
	#else
		return ftp_send_string(control_socket, "200\n", 'a');
	#endif
}


	static int
ftp_stru(int control_socket, const char * arg, char * stru)
{
	/* Checked: 1 */
	/* ## TODO: implement others structures */
	if (strcasecmp(arg, "f") == 0)
		* stru = 'F';
	else if ((strcasecmp(arg, "R") == 0) || (strcasecmp(arg, "P") == 0)) {
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "504 Structure not implemented\n", 'a');
		#else
			return ftp_send_string(control_socket, "504\n", 'a');
		#endif
	} else {
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "500 Illegal STRU command\n", 'a');
		#else
			return ftp_send_string(control_socket, "500\n", 'a');
		#endif
	}
	#ifdef MULTISERV_FTP_VERBOSE
		return ftp_send_string(control_socket, "200 Structure set ok\n", 'a');
	#else
		return ftp_send_string(control_socket, "200\n", 'a');
	#endif
}


	static int
ftp_retr(int control_socket, int * data_socket, const char * arg, char type)
{
	/* Checked: 1 */
	FILE * fp;
	int i;
	fp = fopen(arg, "r");
	if (fp == NULL) {
		close (* data_socket);
		* data_socket = -1;
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "550 No such file\n", 'a');
		#else
			return ftp_send_string(control_socket, "550\n", 'a');
		#endif
	}
	if (type == 'a') {
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "150 Sending ASCII file\n", 'a');
		#else
			ftp_send_string(control_socket, "150\n", 'a');
		#endif
	} else {
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "150 Sending BINARY file\n", 'a');
		#else
			ftp_send_string(control_socket, "150\n", 'a');
		#endif
	}
	do {
		i = fread(multiserv_buffer, 1, MULTISERV_BUFFER_LEN, fp);
		if (i <= 0)
			break;
		if (ftp_send_block(* data_socket, multiserv_buffer, i, type) != 0)
			break;
		/* ## TODO Check for ABORT */
	} while(i == MULTISERV_BUFFER_LEN);
	fclose (fp);
	close (* data_socket);
	* data_socket = -1;
	#ifdef MULTISERV_FTP_VERBOSE
		return ftp_send_string(control_socket, "226 Transfert complete\n", 'a');
	#else
		return ftp_send_string(control_socket, "226\n", 'a');
	#endif
}


	static int
ftp_stor(int control_socket, int * data_socket, const char * arg, char type, char * openmode)
{
	/* Checked: 1 */
	FILE * fp = NULL;
	char unique_filename[9];
	int fd;
	int i;
	/* 
	 * In unique filename store mode (x), we try to create ftp_nnnn,
	 * with nnnn between 0 and 9999.
	 */
	if (* openmode == 'x') {
		for (i = 0; i < 10000; i++) {
			snprintf(unique_filename,9, "ftp_%04d", i);
			fd = open(unique_filename,O_WRONLY | O_CREAT | O_EXCL, 0600);
			if (fd >= 0) {
				fp = fdopen(fd,"w");
				break;
			}
		}
	} else {
		fp = fopen(arg, openmode);
	}
	if (fp == NULL) {
		close (* data_socket);
		* data_socket = -1;
		#ifdef MULTISERV_FTP_VERBOSE
			return ftp_send_string(control_socket, "553 unable to create file\n", 'a');
		#else
			return ftp_send_string(control_socket, "553\n", 'a');
		#endif
	}
	#ifdef MULTISERV_FTP_VERBOSE
		if (type == 'a') {
			ftp_send_string(control_socket, "150 Transfert ASCII file\n", 'a');
		} else {
			ftp_send_string(control_socket, "150 Transfert BINARY file\n", 'a');
		}
	#else
		ftp_send_string(control_socket, "150\n", 'a');
	#endif
	while(1) {
		i = ftp_get_block(* data_socket, multiserv_buffer, MULTISERV_BUFFER_LEN, type);
		if (i <= 0)
			break;
		if (fwrite(multiserv_buffer, 1, i, fp) != i) {
			fclose(fp);
			close (* data_socket);
			* data_socket = -1;
			#ifdef MULTISERV_FTP_VERBOSE
				ftp_send_string(control_socket, "451 Local error (#STOR1)\n", 'a');
			#else
				ftp_send_string(control_socket, "451\n", 'a');
			#endif		
			return (-1);
		}
		/* ## TODO: Check for ABORT */
	}
	fclose (fp);
	close (* data_socket);
	* data_socket = -1;
	#ifdef MULTISERV_FTP_VERBOSE
		if (* openmode == 'x') {
			snprintf(multiserv_buffer, MULTISERV_BUFFER_LEN, "226 Transfert complete (file name:%s)\n", unique_filename);
			return ftp_send_string(control_socket, multiserv_buffer, 'a');
		}
		return ftp_send_string(control_socket, "226 Transfert complete\n", 'a');
	#else
		return ftp_send_string(control_socket, "226\n", 'a');
	#endif
}

/* ------------------ Utilities routines ---------------------------- */

	static char *
ftp_getcwd(void)
{
	/* Checked: 1 */
	char * pathname = NULL;
	char * new;
	int	size = 0;
	
	do {
		size += 4096;
		new = realloc (pathname, size);
		if (new == NULL) {
			if (pathname != NULL)
				free(pathname);
			return (NULL);
		}
		pathname = new;
	} while (getcwd(pathname, size) == NULL);
	return (pathname);
}


	static int
ftp_check_data_socket(int control_socket, int * data_socket, int passive)
{
	/* Checked: 1 */
	int	new_socket;
	if (* data_socket < 0) {
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "425 Use PORT or PASV first\n", 'a');
		#else
			ftp_send_string(control_socket, "425\n", 'a');
		#endif
		return (-1);
	}
	if (! passive)
		/* if positive, an active socket is still open and connected to client */
		return (0);
	/* But passive socket need to be accept()ed */
	if ((new_socket = accept(* data_socket, NULL, NULL)) < 0) {
		close(* data_socket);
		* data_socket = -1;
		#ifdef MULTISERV_FTP_VERBOSE
			ftp_send_string(control_socket, "451 Local error (#CHECK1)\n", 'a');
		#else
			ftp_send_string(control_socket, "451\n", 'a');
		#endif
		return (-1);
	}
	close(* data_socket);
	* data_socket = new_socket;
	return(0);
}


	static int
ftp_send_string(int sock, const char * string, char type)
{
	/* Checked: 1 */
	return ftp_send_block(sock, string, strlen(string), type);
}

	
	static int
ftp_send_block(int sock, const char * block, size_t size, char type)
{
	/* Checked: 1 */
	int i, j;
	j = 0;
	
	if (type == 'i') {
		/* ## TODO: Check if we need to loop if blocksize > MTU ?  */
		if (send(sock, block, size, 0) == size)
			return (0);
		return (-1);
	}
	if (type == 'a') {
		for (i = 0; i < size; i ++) {
			if (block[i] != '\n')
				continue;
			if (i > j) {
				if (send(sock, block+j, i-j, 0) != (i-j))
					return (-1);
			}
			if (send(sock, "\r\n", 2, 0) != 2)
				return (-1);
			j = i+1;
		}
		if (j < i)
			if (send(sock, block+j, i-j, 0) != (i-j))
				return (-1);
		return (0);
	}
	return (-1);
}


	static int
ftp_get_block(int sock, char * block, size_t size, char type)
{
	/* Checked: 1 */
	int i, j;
	i = recv(sock, block, size-1, 0); /* size-1, so we can access i+1 */
	if (i <= 0)
		return (i);
	if (type == 'i')
		return (i);
	if (type == 'a') {
		for (j = 0; j < i; j ++) {
			/* Ignore \r if it is followed by \n */
			if (block[j]=='\r') {
				if (j == (i - 1)) {
					if (recv(sock, block+j+1, 1, 0) != 1)
						return (j+1);
					if (block[j+1] == '\n') {
						block[j] = '\n';
						return (j + 1);
					}
					return (j + 2);
					
				}
				if (block[j+1] == '\n')
					memmove (block+j, block+j+1, i-j-1);
				i --;
			}
		}
		return (i);
	}
	return (-1);
}


	static int
ftp_wait_string (int sock)
{
	/* Checked: 1 */
	int n;
	do {
		n = recv(sock, multiserv_buffer,MULTISERV_BUFFER_LEN-1, 0);
		if (n <= 0)
			return (-1);
		do {
			multiserv_buffer[n] = '\0';
			n --;
		} while ((n>=0) && (isspace(multiserv_buffer[n])));
	} while (n < 0);
	return (0);
}


#ifndef MULTISERV_FTP_ANONYMOUS_ONLY


#define PASSWD_LINE_LEN 1024

	static int
ftp_check_access(const char * logname, const char * password)
{
	/* Check access, chdir() to user home directory, then setgid() and setuid() */
	/* Return 0 if all operations complete, -1 otherwise */
	FILE * fp;
	char line[PASSWD_LINE_LEN];
	
	int i;
	char * field;
	char * pwd;
	int uid, gid;
	/*
	 * We use directly /etc/passwd and /etc/shadow for authentication.
	 * We target small systems on which PAM is probably not installed.
	 */
	
	fp = fopen("/etc/passwd", "r"); 
	if (fp == NULL)
		return (-1);
	do {
		if (fgets(line, PASSWD_LINE_LEN, fp) == NULL) {
			fclose(fp);
			return (-1);
		}
		
		i = 0;
		field = line;
		while(1) {
			if (line[i] == '\0') {
				fclose(fp);
				return (-1);
			}
			if (line[i] == ':') {
				line[i] = '\0';
				break;
			}
			i ++;
		}
	} while (strcmp(logname, field) != 0);

	fclose(fp);
	i ++;
	field = line+i;
	while(1) {
		if (line[i] == '\0')	return (-1);
		if (line[i] == ':') {
			line[i] = '\0';
			break;
		}
		i ++;
	}
	pwd = malloc(strlen(field) + 1);
	if (pwd == NULL)
		return (-1);
	strcpy(pwd, field);
	
	i ++;
	field = line+i;
	while(1) {
		if (line[i] == '\0') {
			free(pwd);
			return (-1);
		}
		if (line[i] == ':') {
			line[i] = '\0';
			break;
		}
		i++;
	}
	if (sscanf(field, "%d", & uid) != 1) {
		free(pwd);
		return (-1);
	}

	i ++;
	field = line+i;
	while(1) {
		if (line[i] == '\0') {
			free(pwd);
			return (-1);
		}
		if (line[i] == ':') {
			line[i] = '\0';
			break;
		}
		i ++;
	}
	if (sscanf(field, "%d", & gid) != 1) {
		free(pwd);
		return (-1);
	}

	i++;
	/* description */
	while(1) {
		if (line[i] == '\0') {
			free(pwd);
			return (-1);
		}
		if (line[i] == ':') {
			line[i] = '\0';
			break;
		}
		i ++;
	}
	
	i++;
	field = line+i;
	while(1) {
		if (line[i] == '\0') {
			free(pwd);
			return (-1);
		}
		if (line[i] == ':') {
			line[i] = '\0';
			break;
		}
		i ++;
	}
	if (chdir(field) != 0) {
		free (pwd);
		return (-1);
	}
	
	if (strcmp(pwd, "x") == 0) {
		free(pwd);
		fp = fopen("/etc/shadow", "r"); 
		if (fp == NULL)
			return (-1);

		do {
			if (fgets(line, PASSWD_LINE_LEN, fp) == NULL) {
				fclose(fp);
				return (-1);
			}
		
			i = 0;
			field = line;
			while(1) {
				if (line[i] == '\0') {
					fclose(fp);
					return (-1);
				}
				if (line[i] == ':') {
					line[i] = '\0';
					break;
				}
				i ++;
			}
		} while (strcmp(logname, field) != 0);

		fclose(fp);
		i ++;
		field = line+i;
		while(1) {
			if (line[i] == '\0')	return (-1);
			if (line[i] == ':') {
				line[i] = '\0';
				break;
			}
			i ++;
		}
		pwd = malloc(strlen(field) + 1);
		if (pwd == NULL)
			return (-1);
		strcpy(pwd, field);
	}
	
	if (pwd != "")
		if (strcmp(crypt(password, pwd), pwd) != 0)
			return (-1);
	if ((setgid(gid) != 0)
	 || (setuid(uid) != 0))
	 	return (-1);
	return(0);
}

#endif

/*	## TODO: Commands still unimplemented:
		ABOR
		STAT
*/


#endif
